'''
movable.py - base classes for movable / rotatable objects
'''

from pyglet.gl import *
import pyglet

from constants import *

from math import pi

from euclid import Vector3, Quaternion

from utilities import heading_to_angle

from descriptions import HasDesc

# ==

class MovingGroup(pyglet.graphics.Group):
    "draw our vertices translated and rotated by a mutable coordinate system"
    # XXX do we need __eq__? not if we just cache the ones with same parent and mover.
    def __init__(self, parent, mover):
        "uses mover.pos and mover.quat"
        pyglet.graphics.Group.__init__(self, parent) #k assume init arg is parent group
        self.mover = mover # knows position, which we use in set_state
    def set_state(self):
        glPushMatrix() # assume GL_MODELVIEW
        p = self.mover.pos
        glTranslatef(p.x, p.y, p.z)
        q = self.mover.quat
        angle, axis = q.get_angle_axis()
        glRotatef(angle * 180. / pi, *axis )
    def unset_state(self):
        glPopMatrix()

# ==

class Movable(HasDesc): # XXX split into Movable and Rollable?
    """
    Abstract class for things that can move and rotate,
    and have a .shape which is drawn in a batch
    in suitably transformed coordinates (by means of a MovingGroup).

    For self to be visible, subclass __init__ or client code must set
    self.shape and call self.add_to_batch.
    """
    debug = False
    direction_vlist = None

    def __init__(self,
                 initpos = ORIGIN,
                 desc = None ):
        "note: you'll also need to call self.add_to_batch"
        
        self.__desc = desc # used by other methods during __init__

        HasDesc.__init__(self, desc = (desc and desc[0]))

        self.pos = initpos.copy() # XXX store this in a collision body? nah, we are one

        # note: we give all movables enough state for rolling and verlet integration,
        # even though some kinds never use this. (XXX revise this? split out Rollable subclass?)
        
        self.lastpos = self.pos.copy() # for verlet integration -- ASSUMES ZERO INITIAL VELOCITY AND ACCELERATION
        self.lastdt = 1. # for time-corrected verlet integration (initial value can be arbitrary nonzero, since initial V & A is 0)
            # (of course the initial A from the terrain is *not* really 0 -- we'll have to see whether this matters;
            #  if it does, see time_corrected_verlet_newpos for what to do about it)
        self.quat = Quaternion() # rotation (owned and mutable)
        self.velocity = Vector3(0., 0., 0.) # only used when not ENABLE_COLLISIONS (I hope)
        self.ground_speed = 0.
        self.roll_rate = 0.
        self.roll_axis = Vector3(0., 0., 0.)
        
        self.__alter_from_desc()
    
    def __alter_from_desc(self):
        if self.__desc:
            self.pos.x, self.pos.y, self.pos.z = self.__desc[1:4]
            self.lastpos = self.pos.copy()

    def get_desc(self): # Movable
        return (
            HasDesc.get_desc(self),
            self.pos.x, self.pos.y, self.pos.z,
        )
    
    def rotate(self, angle, axis):
        "rotate angle (in radians) around axis"
        ## self.quat.rotate_axis(angle, axis)
        ## self.quat = self.quat * Quaternion.new_rotate_axis(angle, axis) # should be and is equivalent, still wrong
        self.quat = Quaternion.new_rotate_axis(angle, axis) * self.quat
            # in theory this is backwards, but it makes rolling correct! bug in euclid, or in my understanding?

    def translate(self, displacement):
        self.pos += displacement
        self.lastpos += displacement
        
    def add_to_batch(self, batch, group):
        self.mygroup = MovingGroup(group, self) # self is the mover, has .pos and .quat
        self.shape.add_to_batch(batch, self.mygroup)
        ## self.shape.vertex_list # could be used to move the player, but using the Group is easier
        # XXX note: the Text movable doesn't yet have .vertex_list; need Shape interface revision

        if self._should_show_downward_line():
            # XXX refactor: only subclass code initializes the data in this;
            # maybe move entire thing into subclass, but then our _batch methods need modification...
            self.direction_vlist = batch.add(2, GL_LINES, None, 'v3f')

    def _should_show_downward_line(self): # Movable
        return self.debug
    
    def remove_from_batch(self):
        self.shape.vertex_list.delete()
        if self.direction_vlist:
            self.direction_vlist.delete()

    pass

# ==

class MovableWithHeading(Movable):
    _up = UP

    def __init__(self,
                 initpos = ORIGIN,
                 heading = Vector3(1., 0., 0.),
                 desc = None ):
        
        Movable.__init__(self, initpos = initpos, desc = desc)
        
        self.heading = heading.copy()
        self.angle = heading_to_angle(self.heading.z, self.heading.x) # angle of heading (zx plane), radians
        ## print "new movable %r: angle = %r, heading = %r" % (self, self.angle, self.heading)

    def camera_pos(self, zoom):
        zoom = 1.0 / zoom
        return self.pos + (self._up * 2.0 - self.heading * 6.0) * zoom

    def lookat(self, zoom):
        up = self._up
        center = self.pos
        eye = self.camera_pos(zoom)
        gluLookAt(eye.x, eye.y, eye.z,
                  center.x, center.y, center.z,
                  up.x, up.y, up.z, )

    # Q: Should we move attrs/methods here to permit direct control of motion, e.g. control_accel?
    # Will we want to control any movable thing's position by direct control? (change to its pov, then roll it...)
    # A: No -- even if we do, we could decide to temporarily parent it with a new object meant to control its position.
    # (This means that even heading, _up, and camera_pos are suspicious here, but I'll leave them for now. XXX)

    pass
