'''
player.py - classes for moving/rolling objects (player, buddies, enemies)
'''

from pyglet.gl import *
import pyglet
from pyglet.window import key

from constants import *
    # this includes:
    ## ORIGIN = Point3(0.0, 0.0, 0.0) ### should be aspect of world or terrain; has multiple defs, some with int vs float XXX

from movable import MovableWithHeading

from utilities import option_key_pressed

from math import pi, sin, cos

from shapes import Sphere

from euclid import Vector3

from textures import USE_SHADOW_TEXTURE

# ==

INIT_PLAYER_POS = ORIGIN + Vector3(2.0, 1.0, 2.0) # XXX or lower (let collision raise it to floor)
INIT_PLAYER_HEADING = Vector3(1.0, 0.0, 1.0).normalized()

PAUSED_SPEED = 6.0

# radius for physics
PLAYER_INIT_RADIUS = 1.
## ENEMY_INIT_RADIUS = 3. # physics doesn't work, goes up too much
## ENEMY_INIT_RADIUS = 0.3 # physics doesn't work -- wall power-assist is too strong, and it can fall through the lawn itself
ENEMY_INIT_RADIUS = 1. # either that, or use that for physics and a smaller or bigger one only for rendering...

# radius for graphics and inter-player collisions; inverted on every other enemy (kluge)
ENEMY_GRAPHICS_RADIUS = 2.

# ==

class FakeKeyStateHandler(dict):
    def __getitem__(self, key):
        return False

the_FakeKeyStateHandler = FakeKeyStateHandler()

# ==

class Player(MovableWithHeading):
    "class for Players, and superclass for similar things (e.g. buddy and enemy balls)"

    # subclass constants
    other = False # true means not the main character; affects controls
    buddy = False
    enemy = False
    terrain_virtual_slope = 1.0
        # value of terrain.VIRTUAL_SLOPE for each Player's physics and shadow-drawing
    
    # variables
    keyboard = None # should be set in .update (kluge) to a KeyStateHandler
    ENABLE_COLLISIONS = 3 # can be toggled externally for debug; chooses one of several collision models

    trail_color = (200,150,100,255) # pinkish dirt color, by default
    
    def __init__(self,
                 initpos = INIT_PLAYER_POS,
                 paused = False,
                 heading = INIT_PLAYER_HEADING,
                 desc = None):
                
        MovableWithHeading.__init__(self, initpos = initpos, heading = heading, desc = desc)

        self.paused = paused # last minute kluge bugfix -- it ought to be a more global state (eg in the world)
        self.radius, self.collision_radius, self.shape = self._create_shape()

        # flags that request one execution of a command (instance variables)
        self.jump = False
        self._do_add_buddy = False
        self._do_add_enemy = False
        
        ## if not self.other:
        ##     print "terrain-collisions:", self.collision_model_short_phrase()
        return
    
    def _create_shape(self):
        "override in subclasses; returns (terrain radius, collision radius (for other players), shape object)"
        return PLAYER_INIT_RADIUS, PLAYER_INIT_RADIUS, Sphere(PLAYER_INIT_RADIUS)

    def _should_show_downward_line(self): # Player
        # if no shadows, this line is a useful indicator of height
        return self.debug or not USE_SHADOW_TEXTURE

    def control_add_buddy(self): # not sure if this belongs in this class, but uses self.pos...
        self._do_add_buddy = True
    
    def control_add_enemy(self): # only inside class Player for historical reasons, I think
        self._do_add_enemy = True
    
    def time_corrected_verlet_newpos(self, dt, A, simple_friction_dt = 1.0):
        """
        Given dt and total acceleration (from terrain, controls, friction, gravity),
        and an optional velocity multiplier for simple friction,
        return desired new self.pos, without setting it in self.
        (Later, caller should pass it to time_corrected_verlet_update, to set it.)
        """
        # http://www.gamedev.net/reference/programming/features/verlet/
        # A Simple Time-Corrected Verlet Integration Method
        # by Jonathan Dummer 
        #
        # Time-Corrected Verlet:
        # xi+1 = xi + (xi - xi-1) * (dti / dti-1) + a * dti * dti
        # 
        # starting condition derivation (if initial V or A was nonzero):
        # xi - xi-1 = vi-1 * dti-1 + 0.5 * ai-1 * dti-1 * dti-1
        #
        # Note: the derivation of verlet integration assumes the acceleration
        # won't change quickly. In practice, it seems to work fine to let the controls
        # vary it quickly (as I think is well known).
        #
        # Note: in theory, one should do verlet integration on the complete
        # game configuration -- that is, the "position" and "acceleration"
        # would be tuples containing all moving object positions, and we'd do
        # one physics update per frame. As long as the terrain acts on each
        # moving object independently, and they don't interact with each other,
        # we can do each one separately. To handle inter-object interactions [NIM],
        # we need to compute those first, store the resulting forces, then use
        # those in this per-player update method. Otherwise this method would
        # get confused by some players' positions and last-positions having been
        # updated too early.
        #
        # The same issue might arise for lawn modifications, but in that case it's
        # perfectly ok if some players modify the lawn seen by other players in the
        # same update iteration (if that's what you want). Just make sure to update
        # the good guys before the bad guys! (Or vice versa, depending on who should
        # have the advantage in the game.) (If you prefer perfect "fairness", then
        # the caller of our update method (which loops over Players) should save up
        # the "what to do to lawn" return values and apply them all at once.)
        
        pos = self.pos
        newpos = ( pos
                   + (pos - self.lastpos) * (dt / self.lastdt) * simple_friction_dt
                   + A * dt * dt
                  )
        return newpos

    def time_corrected_verlet_update(self, dt, newpos):
        self.lastpos = self.pos
        self.pos = newpos
        self.lastdt = dt
        return

    def terrain_accel(self, pos, terrain, debug = False): # XXX make this realistic re terrain, walls, etc # XXX other balls??
        "return the acceleration vector on self (at pos) due to collision force from the terrain"
        # note: caller will also be able to use this to understand friction, if it makes some assumptions
        # note: assume our mass is 1, so we can just compute and return force, not that it matters
        # (since all numeric constants are pretty arbitrary)

        assert self.ENABLE_COLLISIONS

        if self.ENABLE_COLLISIONS == 3:
            # the real potential function, we hope
            return terrain.force_on_ball( pos, self.radius, debug = debug)

        if self.ENABLE_COLLISIONS == 1:
            # initial stub for testing: upward if too low (lower means stronger), and that's all.
            ground_height = 2.5
            equilibrium_dist_under_limit = 0.5
        elif self.ENABLE_COLLISIONS == 2:
            # slightly more realistic stub -- actually notice the terrain height.
            # (not very realistic on slopes! 1. force never points sideways 2. ballsides intersect hillsides)
            ground_height = terrain.terrain_height(pos)
            equilibrium_dist_under_limit = 0.25
        if pos.y < ground_height:
            force = (ground_height - pos.y) * UP * (GRAVITY/equilibrium_dist_under_limit)
            # print "force", force ###
        else:
            force = Vector3(0.0, 0.0, 0.0)
        return force
    
    def control_jump(self):
        "called externally on each jump key press"
        if self.ENABLE_COLLISIONS: # otherwise, no way to get down, though braking stops you high up
            self.jump = True

    def debug_toggle_collisions(self):
        self.ENABLE_COLLISIONS = (1 + self.ENABLE_COLLISIONS) % 4 # 1 false value, 3 true values (2nd/3rd ones are most interesting)
        if not self.other:
            print "terrain-collisions:", self.collision_model_short_phrase()
        self.velocity = Vector3(0,0,0)
        self.lastpos = self.pos.copy()

    def collision_model_short_phrase(self):
        return [  "disabled",
                  "imaginary plane",
                  "real terrain but fake non-conservative potential",
                  "real terrain and potential"] [self.ENABLE_COLLISIONS]

    def compute_control_accel(self, dt, free_fall, terrain_force, fast):
        """
        Return the acceleration vector due to controls (plus non-simple friction).
        (Leave out simple friction (proportional to velocity) and gravity.)
        """
        kb = self.keyboard

        control_accel = Vector3(0.,0.,0.) # modified below

        if not free_fall:
            acceleration = 0.0
            if kb[key.UP]:
                acceleration += ACCELERATION
            if kb[key.DOWN]:
                acceleration -= ACCELERATION
            if fast:
                acceleration *= 2.

            ground_normal = terrain_force.normalized() # or could pass it in

            if acceleration:
                ## direction_of_push = self.heading
                # 090508 improvement (not sure it helps much, but at least it's not worse) --
                # if ground not flat, tilt the plane in which to determine direction_of_push
                right = self.heading.cross(UP)
                direction_of_push = ground_normal.cross(right).normalize()
                    # zero if ground_normal points exactly right, but that should not happen
                
                # XXX make push stronger if velocity is low, i.e. a strong pushoff from standing start?
                
                control_accel += direction_of_push * acceleration
                    ### XXX also needs "rotated tire effect" for better steering (accel transverse to velocity and "up");
                    # friction; ... [see below for more comments on this]

            # XXX jump should depend on ground normal and maybe on "steering"
            if self.jump:
                control_accel += (JUMP_VELOCITY / dt) * (fast and 1.5 or 1)

            # heavy damping when we hit the ground (damp downward V component when touching ground --
            #  we only damp down, not up, otherwise it makes jump too low)
            est_last_velocity = (self.pos - self.lastpos) / self.lastdt # dup code with time_corrected_verlet_newpos
            impact_V = min(0., est_last_velocity.dot(ground_normal)) * ground_normal # component perp to ground (down direction only)
            damp_this_part_out = impact_V * (1. - TERRAIN_IMPACT_DAMPING ** dt)
            control_accel -= damp_this_part_out / dt
            pass
        
        self.jump = False # whether or not free_fall

        ### XXX TODO: non-simple friction (e.g. complex braking, steering == "rotated tire effect").
            # XXX if speed small, set to zero and also "hold still"? (parking brake on slope, and full stopping)

        return control_accel

    def compute_simple_friction(self, free_fall, braking, dt):
        "return the velocity multiplier per second (1.0 for no friction, 0.0 for complete frozen stop)"
        # XXX depend on ground_speed? BUG: ok for ground speed, not up/down damping
        if free_fall:
            return 1.0
        else:
            if braking:
                return BRAKING_FRICTION ** dt
            else:
                return TERRAIN_FRICTION ** dt

    paused = False
    
    def pause_and_print_player_pos(self):
        self.paused = not self.paused
        if not self.other:
            if self.paused:
                print "pausing with player at %r" % self.pos
            else:
                print "resuming"
    
    def update(self, dt, options, terrain, kb, force_from_others = Vector3(0., 0., 0.) ):
        assert dt
        # note: kb can be false even if it's been initialized! It's a dict-like object. Once it's been used,
        # it seems not to be false ever again (guess, not confirmed -- maybe it has explicit keys with False values then?).
        if kb is None:
            print "update with no keyboard -- should never happen"
            return

        ## real_kb = kb # some controls work even for enemies ## none at the moment, though

        controllable_by_kb = (not self.enemy) # i.e. main player or buddy
        if not controllable_by_kb:
            kb = the_FakeKeyStateHandler # kluge?

        self.keyboard = kb # kluge, for access by our submethods
        
        ## fast = kb[key.F] # F = fast (not very useful -- hard to remember, nonintuitive to use)
        # note: key.F is now for toggling fullscreen mode. [old meaning of fast is now hardcoded always on in constants.]
        ## fast = False
        fast = option_key_pressed(kb) # affects many things below: turn speed, accel strength, jump strength

        if self.pos.y < terrain.fallout_height:
            # note: we let this happen to anybody, but intent of game physics
            # is that it can only happen to enemies; physics bugs for long timesteps
            # (e.g. 1 sec, if you create lots of buddies on slow machine) can do it to
            # player/buddies too; world treats player/buddy/enemies differently re fallout
            return dict(I_fell_out = True)

        turn = 0.0
        if kb[key.LEFT]:
            turn += TURN_SPEED
        if kb[key.RIGHT]:
            turn -= TURN_SPEED
        if fast:
            turn *= 1.5
        turn *= dt # number of degrees to rotate heading by
        
        if turn:
            # change heading (controls camera, and direction of control-acceleration)
            self.angle += turn * pi / 180. # note: this is in "zx plane", not "xz plane", thus cos,sin is for z,x respectively:
            self.heading = Vector3(sin(self.angle), 0.0, cos(self.angle)) # assert UP is +y
            ## print "%r changed angle %r heading %r" % (self, self.angle, self.heading)

        if not self.paused:
            # pause helps with screenshots; we let turning (above) still work
            # when paused, as well as lawn-modification (after this 'if')
        
            braking = kb[key.B] # used below (when not free_fall)
                # XXX braking sound (when actually used)

            # last minute kluge -- per-Player terrain.VIRTUAL_SLOPE (so enemies can fall out)
            _old_VIRTUAL_SLOPE = terrain.VIRTUAL_SLOPE
            terrain.VIRTUAL_SLOPE = self.terrain_virtual_slope # different for enemies

            # terrain collision, gravity, position integration
            # part 1: compute A_ter and free_fall
            if self.ENABLE_COLLISIONS: # can be toggled as debug setting (key.Y)
                self.velocity = None # make sure this obsolete attr is not used in this case
                pos = self.pos
                A_ter = self.terrain_accel(pos, terrain) # accel due to terrain contact force
                free_fall = (abs(A_ter) == 0.) # future: might differ if we bounce off "wall" -- not sure
                    # if false, we're touching terrain (in general, intersecting it)
            else:
                # old code (part 1), for debugging
                A_ter = self._up
                free_fall = False

            terrain.VIRTUAL_SLOPE = _old_VIRTUAL_SLOPE

            # part 2: compute things from A_ter
            ground_normal = A_ter.normalized() # used below
            terrain_force = A_ter # assume self's mass is 1
            control_accel = self.compute_control_accel(dt, free_fall, terrain_force, fast) # includes non-simple friction
            simple_friction_dt = self.compute_simple_friction(free_fall, braking, dt)

            mass = 1.0
            if self.enemy and self.collision_radius < 1.0:
                # small enemy has less mass, responds more to force (no effect on terrain physics though)
                mass = self.collision_radius # correct would be collision_radius ** 3 but that seems too extreme
            control_accel += force_from_others / mass
            
            # part 3: gravity, position integration (with simple_friction)
            if self.ENABLE_COLLISIONS:
                control_accel -= self._up * GRAVITY
                A = A_ter + control_accel # total accel
                newpos = self.time_corrected_verlet_newpos(dt, A, simple_friction_dt)
                displacement = newpos - self.pos # used below
                ## if abs(displacement) > 0.5:                
                ##     print "moved", abs(displacement) # rare but happens, but very rare to be more than 1.0
                self.time_corrected_verlet_update(dt, newpos) # sets self.pos and some other variables
            else:
                # no gravity in this case (since no ground to push up when it gets too low)
                self.velocity += control_accel * dt
                self.velocity *= simple_friction_dt
                displacement = self.velocity * dt
                self.pos += displacement

            # ball-rolling appearance        
            if not free_fall:
                # change axis and rate of ball-rolling to match motion along ground
                # (visually, I often think it's rolling too fast, but AFAICT it's accurate,
                #  e.g. when I compare ball surface features with lawn wireframe features.
                #  Would a shadow help?? Actually I just fixed a bug in rolling, and maybe
                #  this has fixed that impression... so maybe i was right.)
                
                ground_displacement = displacement - displacement.dot(ground_normal) * ground_normal
                direction_along_ground = ground_displacement.normalized()
                self.roll_axis = ground_normal.cross(direction_along_ground)
                ground_distance = abs(ground_displacement)
                roll_angle = ground_distance / self.radius # in radians -- this value makes it appear to roll on the surface
                self.roll_rate = roll_angle / dt
            else:
                # roll axis is unchanged, roll rate gradually reduced as if by air friction (for testing, use .3 for not so gradually)
                # XXX might be better to change axis/rate to indicate desired acceleration,
                # as if preparing to do that once touching the ground; not yet, since its control on ground doesn't work that way
                self.roll_rate *= (.9)**dt
                ground_distance = 0.0 # needed below
            
            roll_angle = self.roll_rate * dt
            if abs(self.roll_axis):
                self.rotate(roll_angle, self.roll_axis)

            # speed indicator
            if 0:
                ground_speed = ground_distance / dt
                if int(ground_speed*100) != int(self.ground_speed*100):
                    print ground_speed, "\r",
                    import sys
                    sys.stdout.flush()
                self.ground_speed = ground_speed

            # debugging indicators
            if self.direction_vlist:
                pos = self.pos
                ## end = self.pos + axis * 3.0 # axis end
                end = self.pos - self._up * 10.0
                self.direction_vlist.vertices[:] = (pos.x, pos.y, pos.z, end.x, end.y, end.z)

        else:
            # motion control when paused (new feature 090512, experimental; useful for construction)
            move_forward = move_up = 0.0
            if kb[key.UP]:
                move_forward += 1.0
            if kb[key.DOWN]:
                move_forward -= 1.0
            if kb[key.PAGEUP]:
                move_up += 1.0
            if kb[key.PAGEDOWN]:
                move_up -= 1.0 # XXX issue: player can "fall out" if gets too low!
            if fast:
                move_forward *= 3
                move_up *= 3
            if move_forward or move_up:
                displacement = (move_forward * self.heading + move_up * UP) * PAUSED_SPEED * dt
                # XXX should we also add vertical displacement when terrain height changes??
                self.translate(displacement)
            

        # tell caller how to affect world, by return value # XXX depend on free_fall?
        if self.enemy:
            recolor = options.enable_colortrails and self.trail_color ## we used to use real_kb[key.C] here
        else:
            recolor = ( options.goodguy_coloring and ## we used to use kb[key.G] here
                        {'dirt': self.trail_color,
                         'grass': LAWN_COLOR} [options.goodguy_coloring] )
        what = dict(
            make_hill = kb[key.H],
            make_depression = kb[key.D], # "dig"
            stronger = option_key_pressed(kb),
                ### REVIEW whether to use some other mod key for 'stronger', or a mod key at all
            recolor = recolor
         )
        if self._do_add_buddy:
            self._do_add_buddy = False
            where = self.pos + self.heading * self.new_object_distance(10,20)
            what['add_buddy'] = where

        if self._do_add_enemy:
            self._do_add_enemy = False
            where = self.pos + self.heading * self.new_object_distance(15,30) + \
                    UP * (ENEMY_INIT_RADIUS - self.radius)
            what['add_enemy'] = where
        
        return what

    def new_object_distance(self, small, large):
        # avoid overlapping objects, they blast each other away
        import random
        dist = random.random() * (large - small) + small
        return dist
        
    def get_help_text(self): # XXX CALL ME
        # commented out stuff in the string is NIM; whole thing may not be up to date, eg lacks F8
        return """
            acceleration - arrow keys # or AWSD
            brake - B
            make hill - H
            make depression - D # or M for "mow"??  no, that'll be "mole"...
            color - C
            # jump - space
            # look around - mouse motion
            zoom - mouse wheel # , or 1 or 2
        """

    pass

# ==

class Buddy(Player):
    other = True
    buddy = True

# ==

n_enemies = 0

ENEMY_COLORS = [ # used for trails and shapes
    (200,100,100,255), # pink?? sort of
    (200,200,0,255), # yellow
    (0,128,128,255), # dark bluish green
    (100,100,200,255), # light blue??
    (200,50,200,255), # purple?
 ]

class Enemy(Player):
    
    __save_role = 1
    other = True
    enemy = True
    terrain_virtual_slope = -0.5 # shallow enough to see them roll down the out-of-bounds slope

    def __init__(self, *args, **kws):
        desc = kws.pop('desc', None)
        if desc and not isinstance(desc[0], type(())):
            # BACKWARDS COMPATIBILITY KLUGE: turn old to new; comparison (not the same data in each):
            # old looks like: (27.962279535977906, 0.7732716007754934, 7.0165350025673545, 2.0, (200, 100, 100, 255))
            # new looks like: (((), 21.60647366506274, 0.77304592952210793, 19.875046488339503), 2.0, (200, 100, 100, 255))
            desc = (((),) + desc[0:3], desc[3], desc[4])
        self.__desc = desc
        kws['desc'] = (desc and desc[0])
        Player.__init__(self, *args, **kws)

    def _create_shape(self):
        # LAST MINUTE KLUGE -- do other init effects too
        self.more_init() # sets self.trail_color, used for shape color too
        shape_color = self.trail_color
        global ENEMY_GRAPHICS_RADIUS
        if self.__desc:
            graphics_radius = self.__desc[1]
        else:
            graphics_radius = ENEMY_GRAPHICS_RADIUS
            ENEMY_GRAPHICS_RADIUS = 1. / ENEMY_GRAPHICS_RADIUS # KLUGE
        res = ( ENEMY_INIT_RADIUS,
                graphics_radius,
                Sphere( graphics_radius,
                        rainbow_colored = False, # needed
                        constant_color = shape_color ) )
        return res
    
    def more_init(self):
        global n_enemies
        self.enemy_id = n_enemies # 0, 1, ...
        n_enemies += 1
        if self.__desc:
            color = self.__desc[2]
        else:
            color = ENEMY_COLORS[self.enemy_id % len(ENEMY_COLORS)]
        self.trail_color = color

    def get_desc(self): # Enemy
        "return a value of self.__desc which will recreate our parameters in a new object of the same class"
        # XXX desc format (and many aspects of its usage) is a KLUGE;
        # there is not yet a good way to update its format (except to add fields to it),
        # though altering the "what" representing this class (used in a tuple, (what, desc))
        # might be usable for that
        return (
            Player.get_desc(self),
            self.collision_radius, # note: graphics_radius == collision_radius
            self.trail_color,
        )

    def control_jump(self):
        return
    
    pass
