
from pyglet.gl import *
import pyglet

from euclid import Point3, Plane, Vector3
ORIGIN = Point3(0.0, 0.0, 0.0)
from math import pi, sin, cos

class WarpedRectMesh: # (modified from part of class Torus in pyglet/examples/graphics.py)
    "abstract class for a warped projection of a rectangular mesh"

    ni, nj = 50, 30 # resolution parameters -- number of quads in mesh (overridden in most or all subclasses)
    wrap_i = wrap_j = False # whether to draw quads that bridge the outer edges -- not yet properly implemented re indexing
    _default_rainbow_colored = False
    _default_constant_color = False
    opacity = 64 ## 255 -- doesn't seem to matter unless TRANSPARENT_PLAYERS is set (not surprising)

    def __init__(self, *params, **kws):
##        self.ni = ni # these are one less than slices/inner_slices
##        self.nj = nj
        rainbow_colored = kws.pop('rainbow_colored', None)
        constant_color = kws.pop('constant_color', None)
        assert not kws
        self.params = params # what they are depends on subclass
        if rainbow_colored is None:
            rainbow_colored = self._default_rainbow_colored
        self.rainbow_colored = rainbow_colored
        if constant_color is None:
            constant_color = self._default_constant_color
        self.constant_color = constant_color # note, this is a color 4-tuple of ints, not just a boolean (kluge)
        self.colored = rainbow_colored or constant_color
        # convenience for torus-like things
        self.u_step = 2 * pi / self.ni # big circle == u/i/slices
        self.v_step = 2 * pi / self.nj # small circle == v/j/inner_slices -- halfway around

    def v_n(self, i, j):
        "vertex and normal for position i, j, as 3-tuples"
        ## raise NotImplementedError
        return [i,float(i*i)/self.ni,j], [0,1,0] # normal is wrong, but vertices are useful for testing

    def color(self, i, j):
        ## raise NotImplementedError
        if self.rainbow_colored:
            i = (float(i)/self.ni * 10.) * 2 * pi
            red = int(64 + cos(i) * 127)
            green = int(64 + sin(i) * 127)
            j = (float(j)/self.nj * 5.) * 2 * pi
            blue = int(64 + sin(j) * 127)
            return map(int, [red, green, blue, self.opacity])
        elif self.constant_color:
            return self.constant_color

    def add_to_batch(self, batch, group = None): # refile to super
        
        # Create the vertex and normal arrays, one bigger than our quad-count dimensions.
        vertices = []
        normals = []
        colors = []

        slices = self.ni + 1
        inner_slices = self.nj + 1

        wrap_i = self.wrap_i
        wrap_j = self.wrap_j
        
        for i in range(slices):
            for j in range(inner_slices):
                v, n = self.v_n(i, j)
                vertices.extend(v)
                normals.extend(n)
                if self.colored:
                    colors.extend(self.color(i,j))

        # Create a list of triangle indices, by iterating over lower-left vertices of quads
        indices = []
        for i in range(slices):
            if i == slices - 1 and not wrap_i:
                continue
            for j in range(inner_slices):
                if j == inner_slices - 1 and not wrap_j:
                    continue
                p = i * inner_slices + j # index of lower-left vertex
                indices.extend([p, p + inner_slices, p + inner_slices + 1])
                indices.extend([p, p + inner_slices + 1, p + 1])

        listargs = [
            ('v3f/static', vertices),
            ('n3f/static', normals)
         ]
        
        if self.colored:
            listargs += [('c4B/static', colors)] #k not crashing but not yet showing these colors, probably the setup is wrong
        # if self.textured...
        
        self.vertex_list = batch.add_indexed(len(vertices)//3, 
                                             GL_TRIANGLES,
                                             group,
                                             indices,
                                             *tuple(listargs) )
        return
    pass

class Torus(WarpedRectMesh): # (modified from part of class Torus in pyglet/examples/graphics.py) (not used in game)
    ## params = 1., 0.3 # radius, inner_radius # overridden by __init__
    ni, nj = 50, 30 # resolution parameters
    def v_n(self, i, j): # todo: once per class, for same params (optimization)
        "return [vx,vy,vz], [nx,ny,nz] (as lists for .extend)"
        u = self.u_step * i
        v = self.v_step * j
        
        cos_u = cos(u)
        sin_u = sin(u)

        cos_v = cos(v)
        sin_v = sin(v)

        radius, inner_radius = self.params

        d = (radius + inner_radius * cos_v)
        x = d * cos_u
        y = d * sin_u
        z = inner_radius * sin_v

        nx = cos_u * cos_v
        ny = sin_u * cos_v
        nz = sin_v

        return [x, y, z], [nx, ny, nz]
    
        vertices.extend([x, y, z])
        normals.extend([nx, ny, nz])

    pass

class Sphere(WarpedRectMesh): # (modified from class Torus) (used for Player.shape)
    ## params = (1.,) # radius # overridden by __init__
    ni, nj = 50, 30 # resolution parameters
    # I tried reducing poly count to 25, 25, but rendering speed was not significally affected
    # (testing with windowed, paused, 4 balls total, about 77% cpu).
    _default_rainbow_colored = True
    def v_n(self, i, j): # todo: once per class, for same params (optimization)
        "return [vx,vy,vz], [nx,ny,nz] (as lists for .extend)"
        u = self.u_step * i
        v = self.v_step * j * 0.5 - pi/2.0
        
        cos_u = cos(u)
        sin_u = sin(u)

        cos_v = cos(v)
        sin_v = sin(v)

        (radius,) = self.params

        d = (radius * cos_v)
        x = d * cos_u
        y = d * sin_u
        z = radius * sin_v

        nx = cos_u * cos_v
        ny = sin_u * cos_v
        nz = sin_v

        return [x, y, z], [nx, ny, nz]
    
        vertices.extend([x, y, z])
        normals.extend([nx, ny, nz])

    pass
