'''
terrain.py

terrain mesh, including its collisions (as a potential) and appearance

credit: I read the collide module in Alex Holkner's old pyweek entry "delta-v"
some time ago, and though I didn't refer to it while writing this module
(due to the rules of pyweek), there is probably an influence.
'''

if __name__ == '__main__':
    # for standalone testing, when run from this modules's directory
    import sys
    sys.path.insert(0, "../lib/pyeuclid") # for import euclid
    sys.path.insert(0, "../lib") # for import pyglet

# do all "import *" first, so no risk of masking prior imports (except from prior "import *")
from pyglet.gl import *
from constants import * # e.g. for (radius ** 2 - distance ** 2) ** HARDNESS_EXPONENT * BALL_FORCE_COEFFICIENT

import pyglet

from euclid import Point3, Plane, Vector3

from math import floor, ceil, modf
import random

# ==

def clamp(x, x1, x2): # warning: also defined elsewhere, perhaps differently
    return min(x2, max(x1, x))

class TerrainGroup(pyglet.graphics.Group):
    def __init__(self, parent, terrain):
        pyglet.graphics.Group.__init__(self, parent) #k assume init arg is parent group
        self.terrain = terrain
    def set_state(self):
        if self.terrain.wireframe:
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        else:
            # front fill, back wireframe
            # (for seeing through hills you're under, or walls you're outside,
            #  but not culling them entirely)
            glPolygonMode(GL_FRONT, GL_FILL)
            glPolygonMode(GL_BACK, GL_LINE)
            # simpler value: glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
    def unset_state(self):
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL) # kluge
        
# ==

class Triangle3:
    def __init__(self, *points):
        self.points = points
        p = Plane(*points) # XXX: optimize this away, and remove assumption
        ## self.plane = p
        self.n = p.n # assume this points "out" of terrain
        self.k = p.k
    def hit_sphere(self, center, radius): ### XXX PROBABLY WON'T BE USED
        "return None, or an offset to push sphere just outside of terrain"
        n = self.n
        dist = n.dot(center) - self.k # of center from plane
        if dist >= radius:
            return None
        def hitvertex(pi):
            "it hits point pi only"
            v = center - pi
            av = abs(v)
            if av >= radius:
                return None
            return v * (1.0 - av/radius) ### likely bug: can point backwards re self.n
        def hitedge(pi,pj):
            "it hits edge from pi to pj"
            v = center - pi
            unwanted = v * v.dot((pj-pi).normalize())
            v -= unwanted
            av = abs(v)
            if av >= radius:
                return None
            return v * (1.0 - av/radius) ### likely bug: can point backwards re self.n
        p1, p2, p3 = self.points
##        for pi,pj in ((p1,p2),(p2,p3),(p3,p1)):
##            if (pj-pi).cross(n).dot(center-pi) > 0.0:
##                return # outside edge pj-pi of triangle -- # wrong -- needs to count if it's too close in 3d to edge!
        if (p2-p1).cross(n).dot(center-p1) > 0.0:
            # outside p2-p1
            if (p3-p2).cross(n).dot(center-p2) > 0.0:
                # outside p3-p2
                res = hitvertex(p2)
            else:
                # inside p3-p2
                if (p1-p3).cross(n).dot(center-p3) > 0.0:
                    # outside p1-p3
                    res = hitvertex(p1)
                else:
                    # inside p1-p3
                    res = hitedge(p1,p2)
        else:
            # inside p2-p1
            if (p3-p2).cross(n).dot(center-p2) > 0.0:
                # outside p3-p2
                if (p1-p3).cross(n).dot(center-p3) > 0.0:
                    # outside p1-p3
                    res = hitvertex(p3)
                else:
                    # inside p1-p3
                    res = hitedge(p2,p3)
            else:
                # inside p3-p2
                if (p1-p3).cross(n).dot(center-p3) > 0.0:
                    # outside p1-p3
                    res = hitedge(p3,p1)
                else:
                    # inside p1-p3 -- inside triangle
                    res = n * (radius - dist)
        return res
    def draw_SLOW(self, tex_coord_func = None):
        "slow -- just for testing or for drawing very few triangles"
        glBegin(GL_TRIANGLES)
        for p in self.points: # assume in correct order for drawing
            # assume color set externally, and normal not needed
            # XXX texture coordinate?
            if tex_coord_func:
                u,v = tex_coord_func(p)
                    # note: in practice, that func doesn't know point,
                    # so it can only be written sensibly if it's linear in x and z;
                    # that's ok for our current use
                glTexCoord2f(u,v)
            glVertex3f(*p)
        glEnd()
    pass

# ==

class Terrain:
    """
    Holds a regular mesh of quads for a patch of terrain, in the xz plane.
    
    WARNING: the rendering, heading/angle calcs, etc, use the zx plane (not xz plane),
    in the sense that z is forward-right and x is forward-left (and y is up),
    so the xyz system is right-handed. But the heightmap array is an "x,z matrix"
    in the sense that we say data[xd][zd] and let zd be the inner loop variable.
    """
    wireframe = False
    _copy_colors_from_this = None
    def __init__(self, origin, vx, vz, nx, nz, noise = 0.0, copy_from = None):
        if copy_from:
            # for now, used only when loading from a file;
            # we handle most of the data in copy_from, but caller handles some of it
            nx = copy_from.nx
            nz = copy_from.nz
        self.origin = origin
        self.fallout_height = origin.y - 20. # only used externally
        self.VIRTUAL_SLOPE = 1.
            # last-minute KLUGE -- VIRTUAL_SLOPE is dynamically externally changed, so each player can have their own
        self.vx = vx # XXX might assume parallel to X axis
        self.vz = vz # ditto
        self.nx = nx
        self.nz = nz
        self.scale = abs(self.vx)
        assert abs(vx) == abs(vz) # assume same scale for vz (hardcoded into every use of self.scale)
        if copy_from:
            self.data = copy_from.data # ok to reuse data rather than actually copy it, for now
            self._copy_colors_from_this = copy_from.colors #  copy them later; this might be None, that's ok
        else:
            # initialize self.data (but not colors)
            if not noise:
                self.data = [[0.0] * (nz+1)] * (nx+1)
            else:
                edging_width = min(nz,nx) * 0.1
                pyramid_radius = 3
                central_hill = (nx/2 + pyramid_radius + 2, nz/2 - pyramid_radius - 2)
                central_dip =  (nx/2 - pyramid_radius - 2, nz/2 + pyramid_radius + 2)
                def pd(special_point, slope, xd, zd):
                    # XXX not sure I like these central warts... maybe replace with gentle-dip/hill "symbols"? (like arrows)
                    x = abs(xd - special_point[0])
                    z = abs(zd - special_point[1])
                    return slope * max(0, pyramid_radius - max(x,z))
                def initheight(xd,zd):
                    basic = random.random() * noise
                    extra_height = 0.
                    edge_distance = min(xd,zd,nx-xd,nz-zd)
                    if edge_distance < edging_width:
                        basic *= 4
                        extra_height += (edging_width - edge_distance) * 0.125 # edging has gentle slope, also noisier
                    special = ( pd(central_hill, 0.5, xd, zd) + # XXX would look better if some edge quads' triangulation was reversed!
                                pd(central_dip, -0.5, xd, zd) )
                    return basic + extra_height + special
                self.data = [ [ initheight(xd,zd)
                                for zd in range(nz+1) ]
                              for xd in range(nx+1) ]

    def save_in_file_object(self, file):
        "note: doesn't flush or close; that's up to caller"
        file.write("# outlawn terrain file, format-version = 090509\n")
            # WARNING: this format is partly defined in the caller, which writes more fields! XXX @@@
            # older formats: 090505 didn't have indented continuation lines;
            # 090507 had old form of enemy record (no per-superclass-nesting of param tuples, added in 090509)
        # see files.py for reading code
        file.write("size = %d, %d\n" % (self.nx, self.nz))
        
        file.write("heights = [\n")
        rowformat = " [" + ", ".join(["%.3f"] * len(self.data[0])) + "],\n" # must be indented
        for row in self.data:
            file.write( rowformat % tuple(row) )
        file.write(" ]\n") # must be indented
            # XXX TODO:
            # - (actually, we should probably use an efficient binary format instead!)
            # - (or, an existing text format, like an .obj file?? or JSON?)
        
        colordata = list(self.vertex_list.colors) # 4 unsigned bytes per vertex
        file.write("colors = %r\n" % colordata)
        file.write("# end terrain section\n") # but caller might write other stuff after this (todo: clean up @@@)
        
    def debug_toggle_wireframe(self):
        self.wireframe = not self.wireframe
        
    def meshpoint(self, xd, zd):
        "return an actual meshpoint's 3d location"
        # warning: dup code with add_to_batch
        y = self.data[xd][zd]
        s = self.scale
        return self.origin + Vector3(xd*s,y,zd*s)

    def meshpoint_virtual(self, xd, zd):
        """
        Return a virtual meshpoint's 3d location.
        It can be virtual in two ways:
        - fractional coordinates (we'll interpolate within the triangles)
        - out of range coordinates (we'll continue the mesh in some manner
           sensible for the game -- for now, with rising 45-degree hills
           outside).
        It's also legal to pass integral coordinates as either ints or floats.
        (This might be taken advantage of by recursive calls for handling
         fractional coordinates.)
        """
        # note: using this in an inner loop is pretty inefficient compared
        # to unrolling just the fractional part, but it's simplest to code.
        
        nx = self.nx
        nz = self.nz
        m = self.meshpoint # note: this *can't* tolerate float coords
        
        # first handle virtual points, since for this we treat integral &
        # fractional coords the same.
        slide = 0 # how far we slide down the virtual slope to reach real ground
        slidevec = Vector3(0,0,0) # how we slid horizontally (adding this was a major bugfix re fallout!)
        if xd < 0:
            slide += -xd
            slidevec.x += -xd
            xd = 0
        elif xd > nx:
            slide += xd - nx
            slidevec.x -= xd - nx
            xd = nx
        if zd < 0:
            slide += -zd
            slidevec.z += -zd
            zd = 0
        elif zd > nz:
            slide += zd - nz
            slidevec.z -= zd - nz
            zd = nz
        if slide:
            # both coords are now non-virtual, but if only one of them slid,
            # then the other still might be fractional, so for simplicity,
            # just do a recursive call. Note re slidevec: we have to restore
            # x and z from the slide, not only y!
            return self.meshpoint_virtual(xd, zd) + \
                   UP * (slide * self.VIRTUAL_SLOPE) - slidevec

        # now we know it's a real point, but might be fractional.
        # Get fractional and integral parts (both as floats).
        # (We know xfrac is nonnegative only because xd is. Same for zd.)
        xfrac, xfloor = modf(xd)
        zfrac, zfloor = modf(zd)
        xi = int(xfloor)
        zi = int(zfloor)

        # various cases depending on which vertex or edge or triangle we're in
        def interp(r0, r1, weight):
            return Point3(*tuple(weight * r1 + (1. - weight) * r0))
        def interp2(r0, ra, wa, rb, wb):
            return Point3(*tuple(rb * wb + ra * wa + (1. - wa - wb) * r0))
        if xfrac:
            if zfrac:
                # interior of quad
                where = xfrac + zfrac
                if where == 1.0:
                    # on diagonal that separates the triangles
                    return interp( m(xi+1,zi), m(xi,zi+1), zfrac)
                elif where < 1.0:
                    # interior of bottom triangle
                    return interp2( m(xi,zi), m(xi+1,zi), xfrac, m(xi,zi+1), zfrac)
                else:
                    # interior of top triangle
                    return interp2( m(xi+1, zi+1), m(xi,zi+1), 1. - xfrac, m(xi+1,zi), 1. - zfrac)
            else:
                # zd is integral
                return interp( m(xi,zi), m(xi+1,zi), xfrac)
        else:
            # xd is integral
            if zfrac:
                return interp( m(xi,zi), m(xi,zi+1), zfrac)
            else:
                # both are integral
                return m(xi,zi)
        pass

    def exact_mesh_coords(self, point):
        "return xd, zd for point, without forcing them to be integral or in-range"
        x,z = point.x, point.z # not using point.y, by definition
        s = self.scale
        # someday, fix bug: also use self.origin
        return x/s, z/s
    
    def meshrect_under_noclamp(self, point, radius):
        """
        return xi, xj, zi, zj, which are inclusive ranges of meshpoint indices (xd, zd)
        for meshpoints touching quads under the circle (or rect) with given center point
        and radius, on an imaginary mesh of infinite size in all (2d) directions.
        """
        x,z = point.x, point.z # not using point.y, by definition
        s = self.scale
        # someday, fix bug: also use self.origin
        xi,xj = int(floor((x-radius)/s)), int(ceil((x+radius)/s))
        zi,zj = int(floor((z-radius)/s)), int(ceil((z+radius)/s))
        return xi, xj, zi, zj

    def near_meshpoints(self, point, radius, K, debug = False):
        """
        Yield the virtual meshpoints that are either inside the ball
        (point, radius) or over it. (These are the ones that might
        apply a force to it.)
        """
        xi, xj, zi, zj = self.meshrect_under_noclamp(point, radius)
        rad2 = radius * radius
        # this won't work since range args must be ints:
        ## step = 1./K
        ## inclusive = step * 0.5
        ## ... range(xi, xj + inclusive, step)
        Kf = float(K)
        def frange(xi, xj):
            return [x / Kf for x in range(xi * K, xj * K + 1)]
        bx,by,bz = point
        for xd in frange(xi, xj):
            for zd in frange(zi, zj):
                p = self.meshpoint_virtual(xd, zd)
                if debug:
                    print "examining virtual meshpoint", p
                x,y,z = p
                planedist2 = (x-bx)**2 + (z-bz)**2
                in_xz_by_2 = rad2 - planedist2
                if in_xz_by_2 >= 0: # i.e. if planedist2 <= rad2:
                    dy = p.y - point.y # positive if p is higher than center of ball
                    if dy >= 0 or dy ** 2 <=  in_xz_by_2:
                        yield p # XXX in a tuple with some of this stuff we computed?
        return

    def force_on_ball(self, point, radius, debug = False):
        "(using our most realistic terrain potential function)"
        # out_of_bounds = False
        if point.x < -1 or point.x > self.nx + 1 or \
           point.z < -1 or point.z > self.nz + 1 :
            # out_of_bounds = True # not yet used, in theory should not be needed
            if 0: # temporary debug code
                print
                print "fyi out of bounds at %r" % point
                debug = True
        points = self.near_meshpoints(point, radius, MESH_DENSITY, debug = debug)
        if debug:
            points = list(points) # do debug printing here (fully run the generator)
            print "near_meshpoints returns", points
        bx,by,bz = point
        NEWER_SHIFT_ALGORITHM = False # when True, permits fallout (see comment below)
        INFINITY = 10000000
        if NEWER_SHIFT_ALGORITHM:
            shift_ball_up = INFINITY
        else:
            shift_ball_up = 0.0
        for p in points:
            x,y,z = p
            # look at meshpoint relative to ball -- is it higher than allowed?
            # (i.e. is ball going underground?) if so, we'll compute force on a
            # pretend ball which is higher than this one, and use that in
            # determining force on this one, also pushing up (and over?) to get
            # it aboveground, also never letting the ground push it down
            # (in a net sense, not per-meshpoint). We'll try to conserve energy,
            # but not be as worried about that as about getting it back above
            # ground.
##            dx = x - bx
##            dy = y - by
##            dz = z - bz
            # this is too hard to do correctly right now (based on a ball core,
            # and max permitted ground slope)... so just approximate like this: XXX LIKELY BUG -- probably not good enough
            allowed_mesh_height = by - radius / 2.5 # bottom of ball core
            if not NEWER_SHIFT_ALGORITHM:
                shift_ball_up = max( shift_ball_up, y - allowed_mesh_height )
                # note: test shows this is still needed to prevent fallout, even after
                # the fix in the general bug of noticing any virtual meshpoints
                # when out of bounds (which was fixed using slidevec).
            else:
                # another way -- only if all mesh points agree -- hopefully avoid "depression bounceout artifact"...
                # test results: pro: does fix that; con: permits fallout; removes "power-assist walls" effect, which is fun.
                shift_ball_up_thispointopinion = max( y - allowed_mesh_height, 0.0 )
                shift_ball_up = min( shift_ball_up, shift_ball_up_thispointopinion)
            if debug:
                print "allowed_mesh_height = %r" % allowed_mesh_height
        if NEWER_SHIFT_ALGORITHM:
            if shift_ball_up == INFINITY:
                # no mesh points
                shift_ball_up = 0.0
        assert shift_ball_up >= 0.0
        pretend_pos = point + UP * shift_ball_up
        force = self.force_on_ball_primitive( pretend_pos, radius, debug)
        if force.y < 0.:
            force.y = 0. # never push down
        if shift_ball_up:
            # also push the ball up and over to get it back aboveground
            # (bug: due to approximation, it might not actually be underground!)
            # (note: if it's speeding upwards quickly, we should reduce this -- hopefully not important)
            extra = (UP * GRAVITY * 1.1 + force * min(shift_ball_up, 2.))
            if debug:
                print "shift_ball_up = %r so adding extra %r to force %r, at %r (pretend %r)" % \
                      (shift_ball_up, extra, force, point, pretend_pos)
            force += extra
        return force

    def sum_over_near_meshpoints(self, point, radius, func, K): # XXX obsolete now? not yet...
        # modified, docstring not updated: besides radius test, also use "mesh above ball" test...
        """
        For each meshpoint p (perhaps virtual, perhaps fractional-coordinate)
        within radius of point (in 3d distance), at a mesh density multiplier
        of K (which must be 1 or 2 in present implementation -- easy to accept
        larger ints if necessary), evaluate func(vec, dist2) on the vector
        from p to point and the length**2 (squared) of that vector (a Vector3),
        and return the sum of the results, which should be Vector3's.

        (This is used to sum forces from self's mesh on a ball of the given
        radius centered at point, assuming func encodes a reasonable potential
        for that radius and the physical properties of both ball and mesh.)

        This only makes sense if func would return zero for vectors longer
        than radius, but it is never called for such vectors. Func should
        usually return results in the same direction as the argument vector,
        with the proportionality depending only on its length, and smoothly
        decreasing to 0 at length == radius.
        """
        xi, xj, zi, zj = self.meshrect_under_noclamp(point, radius)
        rad2 = radius * radius
        force = Vector3(0., 0., 0.)
        Kf = float(K)
        def frange(xi, xj):
            return [x / Kf for x in range(xi * K, xj * K + 1)]
        for xd in frange(xi, xj):
            for zd in frange(zi, zj):
                p = self.meshpoint_virtual(xd, zd)
                vec = point - p
                dist2 = vec.magnitude_squared()
                if dist2 < rad2:
                    force += func(vec, dist2)
        return force

    def force_on_ball_primitive(self, point, radius, debug = False): # XXX obs? not yet -- used in force_on_ball
        "(using our most realistic terrain potential function)"
        # XXX note: if the ball can ever get under the terrain, we'll need to fix
        # this with a special case to detect that and push it up, even though
        # that violates the "spherical potential" rule; e.g. a strong upward
        # force whenever the ball core (a smaller ball inside it) intersects
        # an infinite downward cone around a meshpoint, based on the max permitted
        # hill slope (which we might want to enforce, for other reasons too,
        # tho that's NIM now XXXX).
        if debug:
            print
            print "force on ball, point = %r" % (point,)
        
        def func(vec, dist2,
                 # and some precomputed constants (optimization):
                 rad2 = radius * radius
                ):
            ## return vec # stub with correct type, totally wrong otherwise
            # see constants.py for discussion of this formula
            magnitude_divided_by_length = (rad2 - dist2) ** HARDNESS_EXPONENT * BALL_FORCE_COEFFICIENT
            res = vec * magnitude_divided_by_length
            if debug:
                print "vec = %r, dist2 = %r, mag(res) = %r" % (vec, dist2, abs(res))
            return res
        
        return self.sum_over_near_meshpoints( point, radius, func, MESH_DENSITY)
        
    def meshrect_under(self, point, radius):
        """
        return xi, xj, zi, zj, which are inclusive ranges of meshpoint indices (xd, zd)
        for meshpoints touching quads under the circle (or rect) with given center point
        and radius.

        If there are no such points:
        - ideally, one or both of the ranges would be empty (2nd value = 1st value minus 1).
        - but in the present implem, meshpoints at the edge of the 2d mesh are returned.
        """

        xi, xj, zi, zj = self.meshrect_under_noclamp(point, radius)
        
        nx = self.nx
        nz = self.nz
        
        xi = clamp(xi, 0, nx)
        xj = clamp(xj, 0, nx)
        zi = clamp(zi, 0, nz)
        zj = clamp(zj, 0, nz)

        return xi, xj, zi, zj

    def quads_under(self, point, radius, _vertices = False, clamp = True):
        """
        yield the quads (as index pairs xd,zd, same as their lower-left vertices)
        under the circle (actually under its rect bounds)
        """
        if clamp:
            xi, xj, zi, zj = self.meshrect_under(point, radius)
        else:
            xi, xj, zi, zj = self.meshrect_under_noclamp(point, radius)

        # don't bother turning bounding rect into bounding circle, for now

        # meshpoint ranges, xi,xj inclusive, turn into quad ranges, xi,xj exclusive of xj
        
        if _vertices:
            # private option for use by our caller, vertices_under -- make range inclusive
            xj += 1
            zj += 1
        
        for xd in range(xi,xj):
            for zd in range(zi,zj):
                yield xd, zd
        return

    def vertices_under(self, point, radius):
        """
        yield the mesh vertices (as index pairs xd,zd) touching quads
        under the circle (actually under its rect bounds)
        """
        return self.quads_under(point, radius, _vertices = True)

    def terrain_height(self, point):
        "return an approximate terrain height (above y==0) just under point"
##        # stub -- very approximate! just has *something to do* with height, and accurate on flat ground!
##        h = 0.0
##        c = 0
##        for xd,zd in self.vertices_under(point, 0.5): # 0.5 is somewhat arbitrary
##            h += self.data[xd][zd]
##            c += 1
##        assert c
##        return h / c
        # here's a newer version:
        xd, zd = self.exact_mesh_coords(point)
        p = self.meshpoint_virtual(xd, zd)
        return p.y # assume origin.y == 0

    def polys_near(self, point, radius):
        "yield the Triangle3 objects which touch the projection of a 2d circle or rect on self"
        # warning: mostly dup code with polys_near_virtual
        m = self.meshpoint
        for xd, zd in self.quads_under(point, radius):
            e,f,g,h = (
                m( xd,zd),
                m( xd,zd+1),
                m( xd+1,zd),
                m( xd+1,zd+1)
             )
            yield Triangle3(e,f,g)
            yield Triangle3(g,f,h)

    def polys_near_virtual(self, point, radius):
        "slower than polys_near, but handles virtual meshpoints"
        # warning: mostly dup code with polys_near
        m = self.meshpoint_virtual
        for xd, zd in self.quads_under(point, radius, clamp = False):
            e,f,g,h = (
                m( xd,zd),
                m( xd,zd+1),
                m( xd+1,zd),
                m( xd+1,zd+1)
             )
            yield Triangle3(e,f,g)
            yield Triangle3(g,f,h)

    def recolor_vertices_near(self, point, radius, color):
        nzp1 = self.nz+1
        for xd, zd in self.vertices_under(point, radius):
            ### XXX THIS MIGHT BE SLOW compared to using slice assignment for each row, not sure
            
            # can modify self.vertex_list.vertices, tex_coords, colors, indices, normals
            i = xd * nzp1 + zd
            ## i *= 3
            ## self.vertex_list.normals[i:i+3] = (-1.0, 0., 0.) # looks wrong enough to notice (but color depends on view direction)
            i *= 4
            self.vertex_list.colors[i:i+4] = color

    def raise_vertices_near(self, point, radius, heightfunc): # XXX rename -- it can lower as well as raise
        """
        apply heightfunc to the height of all vertices under point (as near as radius);
        it will be passed three args, oldheight and dx/dz relative to point,
        and should return new height (a float). Note that heightfunc might want to take
        dt into account, if its effect on affected vertices is cumulative.
        """
        s = self.scale
        d = self.data
        o = self.origin
        nzp1 = self.nz+1

        # track range of xd, zd at which height changed
        # (could use meshrect_under instead, if desired, but might be bigger)
        low_xd = low_zd =    1000000
        high_xd = high_zd = -1000000

##        radius = 10 ####### for testing        
        for xd, zd in self.vertices_under(point, radius):
            oldheight = d[xd][zd]
            dx = o.x + xd * s - point.x
            dz = o.z + zd * s - point.z
            newheight = heightfunc(oldheight, dx, dz)
##            newheight = 1. ###### testing kluge
            if newheight != oldheight:
##                #print "old->new height", oldheight,newheight###
                d[xd][zd] = newheight
                i = xd * nzp1 + zd
                i *= 3
                pos = self.meshpoint(xd, zd)
                ## pos.y = o.y + newheight
                self.vertex_list.vertices[i:i+3] = (pos.x, o.y + newheight, pos.z) # XXX is o.y used in other places too?
                if xd < low_xd: low_xd = xd
                if xd > high_xd: high_xd = xd
                if zd < low_zd: low_zd = zd
                if zd > high_zd: high_zd = zd
                
        if low_xd > high_xd:
            # no heights changed
            return

        # update normals (could do this later if we might call this more than once per frame)
        xi = low_xd - 1
        xj = high_xd + 1
        zi = low_zd - 1
        zj = high_zd + 1
            # no need to clamp them, update_normals has to do that anyway
        
        self.update_normals( low_xd - 1, high_xd + 1, low_zd - 1, high_zd + 1 )

    def update_normals(self, xi,xj,zi,zj):
        "update normals in an inclusive range"
        nx = self.nx
        nz = self.nz
        nzp1 = nz + 1

        # figure out inclusive range of normals that need updating
        xi = clamp(xi, 0, nx)
        xj = clamp(xj, 0, nx)
        zi = clamp(zi, 0, nz)
        zj = clamp(zj, 0, nz)

        # figure out range of quads whose triangle normals influence that
        qxi = xi-1 # inclusive
        qxj = xj # inclusive (for clamp)
        qzi = zi-1
        qzj = zj
        qxi = clamp(qxi, 0, nx-1) # -1, since last quad index is one less than last vertex index
        qxj = clamp(qxj, 0, nx-1)
        qzi = clamp(qzi, 0, nz-1)
        qzj = clamp(qzj, 0, nz-1)
        qxj += 1 # now exclusive
        qzj += 1

        # XXX use common code here and in the initial normals calc
        # XXX optimize it? seems to be fast enough for now...

        # BEGIN partly-dup code with the initial normals calc (add_to_batch)        
        s = self.scale
        d = self.data
        o = self.origin
        def meshpoint_indexed(xd,zd): # faster than self.meshpoint, I think
            "meshpoint v2"
            y = d[xd][zd]
            return xd, zd, o + Vector3(xd*s,y,zd*s)
        
        normals_accum = [ [ Vector3(0.0, 0.0, 0.0)
                            for zd in range(nz+1) ]
                          for xd in range(nx+1) ] # XXX bigger than necessary, as big as mesh
        counts = [[0.0]*(nz+1)]*(nx+1) # XXX ditto
        
        # for normals, assume it's smooth, so just average them for all 6 neighbor triangles,
        # by accumulating them for all triangles, using Point3's in nested lists to start with.
        def see_triangle(e0,f0,g0):
            "each arg is (xd,zd,point) for some tri vertex"
            e,f,g = e0[2], f0[2], g0[2]
            tri_norm = (f-e).cross(g-f).normalize()
            for xd,zd,_ in (e0,f0,g0):
                normals_accum[xd][zd] += tri_norm
                counts[xd][zd] += 1
                ## indices.append(xd * (nx+1) + zd)
        for xd in range(qxi,qxj):
            for zd in range(qzi,qzj):
                e,f,g,h = (
                    meshpoint_indexed( xd,zd), # XXX inefficient: recalcs each point 4 times
                    meshpoint_indexed( xd,zd+1),
                    meshpoint_indexed( xd+1,zd),
                    meshpoint_indexed( xd+1,zd+1)
                 )
                see_triangle(e,f,g)
                see_triangle(g,f,h)
        # END (most of) dup code
        # now use normals_accum, counts (also dup code)
        for xd in range(xi, xj+1):
            for zd in range(zi, zj+1):
                n = (normals_accum[xd][zd] / counts[xd][zd]).normalize()
                i = xd * nzp1 + zd
                i *= 3
                self.vertex_list.normals[i:i+3] = (n.x, n.y, n.z)

    def collide_sphere(self, point, radius): ### XXX PROBABLY WON'T BE USED; several of these methods could be removed
        '''
        Return pushed_point or None
        '''
        collided = False
        for p in self.polys_near(point, radius * 2.0):
            # the 2.0 is in case early polys push the ball sideways
            push = p.hit_sphere(point, radius)
            if push:
                collided = True
                point = point + push
                    # should work for multiple polys if concavity is not worse
                    # than 90 degrees
        return collided and point or None

    def add_to_batch(self, batch, group):
        # warning: dup code with polys_near
        # warning: very slow, at least due to normals calculation, maybe even otherwise

        group = TerrainGroup(group, self)
        s = self.scale
        d = self.data
        o = self.origin
        nx = self.nx
        nz = self.nz

        indices = [] # triangle indices

        # BEGIN partly-dup code with update_normals
        # (XXX to fix: call update_normals here on entire range, don't add init values for them at all)
        def meshpoint_indexed(xd,zd): # faster than self.meshpoint, I think
            "meshpoint v2"
            y = d[xd][zd]
            return xd, zd, o + Vector3(xd*s,y,zd*s)
        
        normals_accum = [ [ Vector3(0.0, 0.0, 0.0)
                            for zd in range(nz+1) ]
                          for xd in range(nx+1) ]
        counts = [[0.0]*(nz+1)]*(nx+1)
        # for normals, assume it's smooth, so just average them for all 6 neighbor triangles,
        # by accumulating them for all triangles, using Point3's in nested lists to start with.
        def see_triangle(e0,f0,g0):
            "each arg is (xd,zd,point) for some tri vertex"
            e,f,g = e0[2], f0[2], g0[2]
            tri_norm = (f-e).cross(g-f).normalize()
            for xd,zd,_ in (e0,f0,g0):
                normals_accum[xd][zd] += tri_norm
                counts[xd][zd] += 1
                indices.append(xd * (nx+1) + zd)
        for xd in range(0,nx):
            for zd in range(0,nz):
                e,f,g,h = (
                    meshpoint_indexed( xd,zd),
                    meshpoint_indexed( xd,zd+1),
                    meshpoint_indexed( xd+1,zd),
                    meshpoint_indexed( xd+1,zd+1)
                 )
                see_triangle(e,f,g)
                see_triangle(g,f,h)
        # END (most of) dup code with update_normals
        
        vertices = []
        normals = []
        colors = []
        def m(xd,zd):
            "meshpoint"
            y = d[xd][zd]
            return o + Vector3(xd*s,y,zd*s)
        for xd in range(0,nx+1):
            for zd in range(0,nz+1):
                p = m(xd,zd)
                n = (normals_accum[xd][zd] / counts[xd][zd]).normalize()
                vertices.extend([p.x, p.y, p.z])
                normals.extend([n.x,n.y,n.z])
                colors.extend(LAWN_COLOR) # initial lawn color
# all seem correct:
##        print "nx = %r, nz = %r, len(vertices) == %r, len(normals) == %r, len(indices) == %r" % \
##              ( nz, nz, len(vertices), len(normals), len(indices) )
# nx = 10, nz = 10, len(vertices) == 363, len(normals) == 363, len(indices) == 600

        if self._copy_colors_from_this:
            assert len(self._copy_colors_from_this) == len(colors)
            colors = self._copy_colors_from_this
            
        self.vertex_list = batch.add_indexed(len(vertices)//3, 
                                             GL_TRIANGLES,
                                             group,
                                             indices,
                                             ('v3f', vertices), # no longer /static, not speed tested since then
                                             ('n3f', normals),
                                             ('c4B', colors),
                                             )
        # note: self.vertex_list.colors (for example) is of type
        # pyglet.graphics.vertexattribute.c_ubyte_Array_14884 (array length is encoded into the type);
        # list() converts it into (something that prints as) a list of python ints from 0 to 255.
        return

    def draw_subset_SLOW(self, point, radius, virtual = False, tex_coord_func = None):
        if virtual:
            polys = self.polys_near_virtual(point, radius)
        else:
            polys = self.polys_near(point, radius)
        for poly in polys:
            poly.draw_SLOW(tex_coord_func = tex_coord_func)
    pass

# ==

if __name__ == '__main__':
    # test code (there is also some near start of file)
    # (see also a separate test in ../test/graphicstest.py)
    p1 = Point3(0,0,0)
    p2 = Point3(1,1,1)
    v = p2 - p1
    print "vec v =", v
    v.normalize()
    print v
    assert isinstance(v+v, Vector3)
    assert isinstance(p1+v, Point3)
    assert isinstance(v+p1, Point3)
    print "what type is the sum of two points?", p2+p2 # Vector3
    ## assert isinstance(p2+p2, Point3) ### ???
    print "what type is a weighted point?", p2*0.5 # Vector3
    print "what if the weight is 1.0?", p2*1.0 # Vector3

    tt = Triangle3(Point3(0,0,0), Point3(0,0,1), Point3(1,0,0))
    assert tt.hit_sphere(Point3(0.25,0.5,0.25),1)
    assert not tt.hit_sphere(Point3(0.25,1.5,0.25),1) # too high
    assert tt.hit_sphere(Point3(0.5,0.5,0.75),1) # too far to one side -- but too close to that edge anyway! (I guess, not verified!)
    assert tt.hit_sphere(Point3(0.75,0.5,0.5),1) # ... to another side (but ditto)
    assert not tt.hit_sphere(Point3(0.5,1.0,0.75),1) # too far to one side, and too far above too
    print "triangle tests done"

    N = 100
    tr = Terrain( ORIGIN, Vector3(1,0,0), Vector3(0,0,1), N, N, noise = 0.1)
    for y2 in range(8):
        y = y2/2.0
        point = Point3(3.2,y,7.0)
        print y, tr.collide_sphere(point,1) # this result is only checked manually
    print "collide_sphere tests done"

    import pyglet
    b = pyglet.graphics.Batch()
    g = pyglet.graphics.Group()
    import time
    time1 = time.time()
    tr.add_to_batch(b,g)
    time2 = time.time()
    print "tr.add_to_batch test done (%r triangles, %.3f seconds)" % (N*N*2, time2 - time1)
        # (test results from before I fixed bug in initializing normals_accum, and before noise:)
        # N = 50,   5000 triangles, 0.621 seconds
        # N = 100, 20000 triangles, 2.558 seconds
        # N = 200, 80000 triangles, 9.925 seconds [roughly linear]
        # [after bugfix: 80000 triangles, 10.097 seconds]
        #
        # this timing might be tolerable for level loading (though slow),
        # but would be intolerable for in-game continuous update if we're
        # modifying the mesh, e.g. growing a hill on it.
        # That was the goal, so, it needs to be faster (or the goal needs to be revised).
        #
        # Note, I'm assuming that in-place updating would be just as bad,
        # but didn't test that. It seems very likely though.

    def heightfunc_FOR_UNIT_TEST(old, dx, dz): # the name is to stop me from changing the wrong one
        return old + max(0., abs(3.-dx-dz))
    tr.raise_vertices_near( ORIGIN + Vector3(N/2, N/2, N/2), 5., heightfunc_FOR_UNIT_TEST)
    time3 = time.time()
    print "time to raise_vertices_near", time3 - time2
        
